from asn1crypto import core
from asn1crypto.keys import PublicKeyInfo as SubjectPublicKeyInfo
from asn1crypto.x509 import AlgorithmIdentifier
from minikerberos.protocol.asn1_structs import EncryptionKey, Checksum, KerberosTime, PrincipalName, Realm


#### PKINIT ASN1 strructs
#### RFC4556 # https://tools.ietf.org/html/rfc4556 ####

# KerberosV5Spec2 DEFINITIONS EXPLICIT TAGS ::=
TAG = 'explicit'

# class
UNIVERSAL = 0
APPLICATION = 1
CONTEXT = 2

class DHNonce(core.OctetString):
	pass

class AlgorithmIdentifiers(core.SequenceOf):
	_child_spec = AlgorithmIdentifier

class TD_DH_PARAMETERS(core.SequenceOf):
	_child_spec = AlgorithmIdentifier

class ReplyKeyPack(core.Sequence):
	_fields = [
		('replyKey', EncryptionKey, {'tag_type': TAG, 'tag': 0}), 
		('asChecksum', Checksum, {'tag_type': TAG, 'tag': 1}),
		('dhKeyExpiration', KerberosTime, {'tag_type': TAG, 'tag': 2, 'optional': True}),
	]

class KDCDHKeyInfo(core.Sequence):
	_fields = [
		('subjectPublicKey', core.BitString, {'tag_type': TAG, 'tag': 0}), 
		('nonce', core.Integer, {'tag_type': TAG, 'tag': 1}),
		('dhKeyExpiration', KerberosTime, {'tag_type': TAG, 'tag': 2, 'optional': True}),
	]

class DHRepInfo(core.Sequence):
	_fields = [
		('dhSignedData', core.OctetString, {'tag_type': 'implicit', 'tag': 0}), 
		('serverDHNonce', DHNonce, {'tag_type': TAG, 'tag': 1, 'optional': True}),

	]

class PA_PK_AS_REP(core.Choice):
	_alternatives = [
		('dhInfo', DHRepInfo, {'explicit': (CONTEXT,0) }  ),
		('encKeyPack', core.OctetString, {'implicit': (CONTEXT,1) }  ),
	]

class ExternalPrincipalIdentifier(core.Sequence):
	_fields = [
		('subjectName', core.OctetString, {'tag_type': 'implicit', 'tag': 0, 'optional' : True}), 
		('issuerAndSerialNumber', core.OctetString, {'tag_type': 'implicit', 'tag': 1, 'optional' : True}),
		('subjectKeyIdentifier', core.OctetString, {'tag_type': 'implicit', 'tag': 2, 'optional' : True}), 
	]

class ExternalPrincipalIdentifiers(core.SequenceOf):
	_child_spec = ExternalPrincipalIdentifier

class AD_INITIAL_VERIFIED_CAS(core.SequenceOf):
	_child_spec = ExternalPrincipalIdentifier

class KRB5PrincipalName(core.Sequence):
	_fields = [
		('realm', Realm, {'tag_type': TAG, 'tag': 0}), 
		('principalName', PrincipalName, {'tag_type': TAG, 'tag': 1}),
	]

class TD_INVALID_CERTIFICATES(core.SequenceOf):
	_child_spec = ExternalPrincipalIdentifier

class TD_TRUSTED_CERTIFIERS(core.SequenceOf):
	_child_spec = ExternalPrincipalIdentifier

class PKAuthenticator(core.Sequence):
	_fields = [
		('cusec', core.Integer, {'tag_type': TAG, 'tag': 0}), 
		('ctime', KerberosTime, {'tag_type': TAG, 'tag': 1}),
		('nonce', core.Integer, {'tag_type': TAG, 'tag': 2}),
		('paChecksum', core.OctetString, {'tag_type': TAG, 'tag': 3, 'optional': True}),
	]

class AuthPack(core.Sequence):
	_fields = [
		('pkAuthenticator', PKAuthenticator, {'tag_type': TAG, 'tag': 0}), 
		('clientPublicValue', SubjectPublicKeyInfo, {'tag_type': TAG, 'tag': 1, 'optional' : True}),
		('supportedCMSTypes', AlgorithmIdentifiers, {'tag_type': TAG, 'tag': 2, 'optional' : True}), 
		('clientDHNonce', DHNonce, {'tag_type': TAG, 'tag': 3, 'optional' : True}), 

	]

class PA_PK_AS_REQ(core.Sequence):
	_fields = [
		('signedAuthPack', core.OctetString, {'tag_type': 'implicit', 'tag': 0}), 
		('trustedCertifiers', ExternalPrincipalIdentifiers, {'tag_type': TAG, 'tag': 1, 'optional' : True}),
		('kdcPkId', core.OctetString, {'tag_type': 'implicit', 'tag': 2, 'optional' : True}), 
	]


# TODO: figure out what structs these are...

class NameTypeAndValueBMP(core.Sequence):
	_fields = [
		('type', core.ObjectIdentifier),
		('value', core.BMPString),
	]

class Dunno1(core.SetOf):
	_child_spec = NameTypeAndValueBMP

class Dunno2(core.SequenceOf):
	_child_spec = Dunno1

class Info(core.Sequence):
	_fields = [
		('pku2u', core.GeneralString, {'tag_type': TAG, 'tag': 0}),
		('clientInfo', PrincipalName, {'tag_type': TAG, 'tag': 1}),
	]

class CertIssuer(core.Sequence):
	_fields = [
		('data', core.OctetString, {'tag_type': 'implicit', 'tag': 0}), # there is another ASN1 encoded blob here that contains the issuer. Classes X and Y deal with that. No documentation....
	]

class CertIssuers(core.SequenceOf):
	_child_spec = CertIssuer

class MetaData(core.Sequence):
	_fields = [
		('1', CertIssuers, {'tag_type': TAG, 'tag': 0}), 
		('Info', Info, {'tag_type': TAG, 'tag': 1}),
	]


import logging
import asyncio
from minikerberos.common.factory import KerberosClientFactory, kerberos_url_help_epilog
from minikerberos.protocol.external.ticketutil import get_NT_from_PAC

async def get_NT(kerberos_url):
	cu = KerberosClientFactory.from_url(kerberos_url)
	client = cu.get_client()
	tgs, enctgs, key, decticket = await client.U2U()
	results = get_NT_from_PAC(client.pkinit_tkey, decticket)
	for result in results:
		print('%s : %s' % (result[0], result[1]))
	print('Done!')
	return results

def main():
	import argparse
	
	parser = argparse.ArgumentParser(description='Fetches the NT hash for the user. PKI auth required.', formatter_class=argparse.RawDescriptionHelpFormatter, epilog = kerberos_url_help_epilog)
	parser.add_argument('kerberos_url', help='The kerberos target URL. Example: "kerberos+pfx://TEST.corp\Administrator:admin@10.10.10.2/?certdata=test.pfx"')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	
	args = parser.parse_args()
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
	else:
		logging.basicConfig(level=1)
	
	asyncio.run(get_NT(args.kerberos_url))
	
	
if __name__ == '__main__':
	main()


from minikerberos.common.factory import KerberosClientFactory
from minikerberos.common.spn import KerberosSPN
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.kirbi import Kirbi
from minikerberos.protocol.external.ticketutil import get_NT_from_PAC
from minikerberos.protocol.ticketutils import construct_apreq_from_tgs_tgt


import tempfile
import asyncio
import pathlib
import shutil
import os

import pytest 

from .config import *


def test_tgt_pw():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PW)
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None


def test_tgt_pw_incorrect_realm():
	cu = KerberosClientFactory.from_url('kerberos+password://NONEXISTENT\\%s:%s@10.10.10.2' % (KERBEROS_USER, KERBEROS_PASSWORD))
	client = cu.get_client_blocking()
	with pytest.raises(Exception) as e_info:
		client.get_TGT()


def test_tgt_pw_incorrect_user():
	cu = KerberosClientFactory.from_url('kerberos+password://%s\\NOTVALIDUSER:%s@10.10.10.2' % (KERBEROS_DOMAIN, KERBEROS_PASSWORD))
	client = cu.get_client_blocking()
	with pytest.raises(Exception) as e_info:
		client.get_TGT()



def test_tgt_pw_incorrect_pw():
	cu = KerberosClientFactory.from_url('kerberos+password://%s\\%s:blablaThisiswrong12@10.10.10.2' % (KERBEROS_DOMAIN, KERBEROS_USER))
	client = cu.get_client_blocking()
	with pytest.raises(Exception) as e_info:
		client.get_TGT()
	



def test_tgt_rc4():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_RC4)
	cred = cu.get_creds()
	assert len(cred.get_supported_enctypes()) == 2
	assert 23 in cred.get_supported_enctypes(as_int=True)
	assert -128 in cred.get_supported_enctypes(as_int=True)
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None


def test_tgt_aes128():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_AES128)
	cred = cu.get_creds()
	assert len(cred.override_etypes) == 1
	assert cred.override_etypes[0].value == 17
	assert cred.get_supported_enctypes(as_int=True) == [17]
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None


def test_tgt_aes256():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_AES256)
	cred = cu.get_creds()
	assert len(cred.override_etypes) == 1
	assert cred.override_etypes[0].value == 18
	assert cred.get_supported_enctypes(as_int=True) == [18]
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None

# TODO: need keytab to test
#
#def test_tgt_keytab():
#    cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_AES256)
#    cred = cu.get_creds()
#    assert len(cred.override_etypes) == 1
#    assert cred.override_etypes[0].value == 18
#    assert cred.get_supported_enctypes(as_int=True) == [18]
#    client = cu.get_client_blocking()
#    client.get_TGT()
#    assert client.kerberos_TGT is not None
#    assert client.kerberos_TGT_encpart is not None


def test_tgs_direct():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PW)
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None
	spn = KerberosSPN.from_spn(KERBEROS_TGS_SPN)
	tgs, encTGSRepPart, key = client.get_TGS(spn)
	assert tgs is not None
	assert encTGSRepPart is not None
	assert key is not None


def test_tgs_direct_incorrect_spn():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PW)
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None
	spn = KerberosSPN.from_spn('whatis/noserver.%s@%s' % (KERBEROS_DOMAIN, KERBEROS_DOMAIN))
	with pytest.raises(Exception) as e_info:
		tgs, encTGSRepPart, key = client.get_TGS(spn)



def test_tgs_ccache():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PW)
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None
	with tempfile.NamedTemporaryFile() as file:
		client.ccache.to_file(file.name)
		spn = KerberosSPN.from_spn(KERBEROS_TGS_SPN)
		creds = KerberosCredential.from_ccache(file.name, encoding = 'file')
		client2 = cu.get_client_newcred_blocking(creds)
		tgs, encTGSRepPart, key = client2.get_TGS(spn)
		assert tgs is not None
		assert encTGSRepPart is not None
		assert key is not None

def test_tgs_ccache_tgs():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PW)
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None
	with tempfile.NamedTemporaryFile() as file:
		client.ccache.to_file(file.name)
		spn = KerberosSPN.from_spn(KERBEROS_TGS_SPN)
		creds = KerberosCredential.from_ccache(file.name, encoding = 'file')
		client2 = cu.get_client_newcred_blocking(creds)
		tgs, encTGSRepPart, key = client2.get_TGS(spn)
		assert tgs is not None
		assert encTGSRepPart is not None
		assert key is not None
		tgs, encTGSRepPart, key = client2.get_TGS(spn)
		assert tgs is not None
		assert encTGSRepPart is not None
		assert key is not None


def test_tgs_kirbi():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PW)
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None
	kirbi = Kirbi.from_ticketdata(client.kerberos_TGT, client.kerberos_TGT_encpart)
	spn = KerberosSPN.from_spn(KERBEROS_TGS_SPN)
	creds = KerberosCredential.from_kirbi(kirbi, encoding = 'kirbi')
	client2 = cu.get_client_newcred_blocking(creds)
	tgs, encTGSRepPart, key = client2.get_TGS(spn)
	assert tgs is not None
	assert encTGSRepPart is not None
	assert key is not None


def test_delegation_self():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_MACHINE)
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None

	service_spn = KerberosSPN.from_spn(KERBEROS_DELEGATION_SPN_SELF)
	target_user = KerberosSPN.from_upn(KERBEROS_DELEGATION_USER_SELF)

	tgs, encTGSRepPart, key = client.S4U2self(target_user, service_spn)
	assert tgs is not None
	assert encTGSRepPart is not None
	assert key is not None


def test_delegation_proxy():
	#moving certificate to current path...
	current_file_path = pathlib.Path(__file__).parent.absolute()
	pfx_file_path = current_file_path.joinpath('testdata', 'test.pfx')
	try:
		shutil.copy(pfx_file_path, 'a.pfx')
		cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_MACHINE)
		client = cu.get_client_blocking()
		client.get_TGT()
		assert client.kerberos_TGT is not None
		assert client.kerberos_TGT_encpart is not None

		service_spn = KerberosSPN.from_spn(KERBEROS_DELEGATION_SPN_PROXY)
		target_user = KerberosSPN.from_upn(KERBEROS_DELEGATION_USER_PROXY)

		tgs, encTGSRepPart, key = client.getST(target_user, service_spn)
		assert tgs is not None
		assert encTGSRepPart is not None
		assert key is not None
	finally:
		os.remove('a.pfx')


def test_pkinit_pfx():
	#moving certificate to current path...
	current_file_path = pathlib.Path(__file__).parent.absolute()
	cert_file_path = current_file_path.joinpath('testdata', 'test.pfx')
	try:
		shutil.copy(cert_file_path, 'a.pfx')
		cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PKINIT_PFX)
		client = cu.get_client_blocking()
		client.get_TGT()

		assert client.kerberos_TGT is not None
		assert client.kerberos_TGT_encpart is not None
	finally:
		os.remove('a.pfx')


def test_pkinit_pem():
	current_file_path = pathlib.Path(__file__).parent.absolute()
	cert_file_path = current_file_path.joinpath('testdata', 'test.pem')
	key_file_path = current_file_path.joinpath('testdata', 'test.key')
	try:
		shutil.copy(cert_file_path, 'a.pem')
		shutil.copy(key_file_path, 'a.key')
		cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PKINIT_PEM)
		client = cu.get_client_blocking()
		client.get_TGT()

		assert client.kerberos_TGT is not None
		assert client.kerberos_TGT_encpart is not None
	finally:
		os.remove('a.pem')
		os.remove('a.key')


def test_pkinit_u2u():
	#moving certificate to current path...
	current_file_path = pathlib.Path(__file__).parent.absolute()
	cert_file_path = current_file_path.joinpath('testdata', 'test.pfx')
	try:
		shutil.copy(cert_file_path, 'a.pfx')
		cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PKINIT_PFX)
		client = cu.get_client_blocking()
		tgs, enctgs, key, decticket = client.U2U()
		results = get_NT_from_PAC(client.pkinit_tkey, decticket)
		assert len(results) > 0
		
	finally:
		os.remove('a.pfx')

def test_md4():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_RC4MD4)
	cred = cu.get_creds()
	assert len(cred.get_supported_enctypes()) == 1
	assert -128 in cred.get_supported_enctypes(as_int=True)
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None


def test_tgs_direct_asreq():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PW)
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None
	spn = KerberosSPN.from_spn(KERBEROS_TGS_SPN)
	tgs, encTGSRepPart, key = client.get_TGS(spn)
	assert tgs is not None
	assert encTGSRepPart is not None
	assert key is not None
	asreq = construct_apreq_from_tgs_tgt(tgs, key, client.kerberos_TGT, flags = None, seq_number = 0, ap_opts = [], cb_data = None)
	assert asreq is not None

def test_tgs_direct_asreq_1():
	cu = KerberosClientFactory.from_url(KERBEROS_CONN_URL_PW)
	client = cu.get_client_blocking()
	client.get_TGT()
	assert client.kerberos_TGT is not None
	assert client.kerberos_TGT_encpart is not None
	spn = KerberosSPN.from_spn(KERBEROS_TGS_SPN)
	tgs, encTGSRepPart, key = client.get_TGS(spn)
	assert tgs is not None
	assert encTGSRepPart is not None
	assert key is not None
	asreq = client.construct_apreq(tgs, None, key, flags = None, seq_number = 0, ap_opts = [], cb_data = None)
	assert asreq is not None

#
#def test_keytab_1():
#	cu = KerberosClientFactory.from_url('kerberos+keytab://www.test.corp:app2.keytab@10.10.10.2')
#	cred = cu.get_creds()
#	print(str(cred))
#	print(cred.get_supported_enctypes())
#	client = cu.get_client_blocking()
#	client.get_TGT()
#	assert client.kerberos_TGT is not None
#	assert client.kerberos_TGT_encpart is not None
	

def main():
	test_tgs_direct_asreq_1()

if __name__ == '__main__':
	main()


# Description: Tests for the kirbi module

import tempfile
import pathlib
from minikerberos.common.kirbi import Kirbi
from .config import *


def test_load_kirbi():
    for kirbifile in get_testfiles_kirbi():
        kirbi = Kirbi.from_file(kirbifile)
        desc = str(kirbi)
        with tempfile.NamedTemporaryFile() as f:
            kirbi.to_file(f.name)
            kirbi2 = Kirbi.from_file(f.name)
            desc2 = str(kirbi2)
            assert desc == desc2

def test_kirbi_b64():
    for kirbifile in get_testfiles_kirbi():
        kirbi = Kirbi.from_file(kirbifile)
        desc1 = str(kirbi)
        kirbi_b64 = kirbi.to_b64()
        kirbi2 = Kirbi.from_b64(kirbi_b64)
        desc2 = str(kirbi2)
        assert desc1 == desc2

def test_kirbi_hex():
    for kirbifile in get_testfiles_kirbi():
        kirbi = Kirbi.from_file(kirbifile)
        desc1 = str(kirbi)
        kirbi_hex = kirbi.to_hex()
        kirbi2 = Kirbi.from_hex(kirbi_hex)
        desc2 = str(kirbi2)
        assert desc1 == desc2

def test_kirbi_bytes():
    for kirbifile in get_testfiles_kirbi():
        kirbi = Kirbi.from_file(kirbifile)
        desc1 = str(kirbi)
        kirbi_bytes = kirbi.to_bytes()
        kirbi2 = Kirbi.from_bytes(kirbi_bytes)
        desc2 = str(kirbi2)
        assert desc1 == desc2
